
from . import seqio 
from . import stringops
from . import misc
from . import decorators

from .listops import batch_list
from .listops import order_cols



__all__ = ['stringops', 'seqio', 'batch_list', 'order_cols', 'decorators']




import re 
import xml.etree.ElementTree as ET


def get_variation_archive_info(s, fh, query_end):
    info = []
    slist = [s]
    while True:
        s = fh.readline()
        slist.append(s)
        if (not s) or (re.search(query_end, s)):
            break
            
    root = ET.fromstringlist(slist)
    
    accession = root.get("Accession")
    
    # find all variant infos 
    for hgvs_record in root.findall('.//HGVS'):
        if hgvs_record.get("Type") == "coding":
            
            # only consider missense mutations 
            mut_type = hgvs_record.find('MolecularConsequence')
            if mut_type is None or mut_type.get('Type') != 'missense variant':
                continue 
            
            # get the protein expression record 
            protein_elem = hgvs_record.find('ProteinExpression')
            nuc_elem = hgvs_record.find("NucleotideExpression")
            
            if protein_elem is None:
                continue
            
            mane_select = bool(nuc_elem.get("MANESelect", False))
            
            # record the changes 
            info.append({'seq_accession': protein_elem.get('sequenceAccessionVersion'),
                         'change': protein_elem.get('change'),
                         'mane_select': mane_select})        
            
    return accession, info




#### Utilities for batching a computation. ####
#### taken from: https://github.com/aprkh/aprkh_utils/blob/main/decorators/batch_utils.py ####

from functools import wraps 
from functools import reduce 


# assumes that `func` is a function that operates on a list
class BatchComp():

    def __init__(self, block_size=10000):
        self.block_size = block_size

    def __call__(self, func):

        @wraps(func)
        def inner(data, *args, **kwargs):
            nobs = len(data)
            lo = 0
            niter = (nobs // self.block_size) + (0 if nobs % self.block_size == 0 else 1)
            results = [None for _ in range(niter)]
            ct = 0
            while lo < nobs:
                hi = min(lo + self.block_size, nobs)
                print(f"computing results for observations {lo} through {hi-1}...", end=' ')            
                results[ct] = func(data[lo:hi], *args, **kwargs)
                print("done!")
                lo = hi
                ct += 1
            return results

        return inner

    def set_block_size(self, n):
        assert isinstance(n, int)
        assert n > 1
        self.block_size = n


def reduce_dict(dict_list):
    return reduce(update_dict, dict_list)


def update_dict(d1, d2):
    d1.update(d2)
    return d1
    


# %%
import pandas as pd
import numpy as np 
import os 
import sys
import logomaker as lm
import matplotlib.pyplot as plt
import logging
from tqdm import tqdm
from IDmapping import main as IDmapping_main
logger = logging.getLogger(__name__)
from pythonjsonlogger import jsonlogger
import json 
from collections import Counter
os.chdir(os.path.dirname(os.path.abspath(__file__)))
print(f'setting working directory to: {(os.path.dirname(os.path.abspath(__file__)))}')

def get_background_distribution(substrates: pd.Index , position=None):
    symbols = ['A', 'R', 'N', 'D', 'C', 'E', 'Q', 'G', 'H', 'I', 'L', 'K', 'M', 'F', 'P', 'S', 'T', 'W', 'Y', 'V']
    bkgrdCount = Counter(dict.fromkeys(symbols, int(0)))
    if position is not None:
        counts = substrates.str.upper().str[position].value_counts().astype(int).to_dict()
        bkgrdCount.update(counts)
    else:
        for i in range(15):
            currcounts = Counter(substrates.str.upper().str[i].value_counts().astype(int).to_dict())
            bkgrdCount = bkgrdCount + currcounts
    return dict(bkgrdCount)

def make_bkgrd_matrix(distributions:dict[dict], position_specific=True):
    dist = {int(pos)-7: distributions[pos] for pos in distributions if pos.isdigit()}
    matrix = pd.DataFrame.from_records(dist).T
    matrix.replace(0, np.NaN, inplace=True)
    bkgrd_probs = get_probability(matrix)
    bkgrd_probs.drop(columns=[x for x in bkgrd_probs if x not in symbols], inplace=True)
    return bkgrd_probs


def assertPhosphosite(kinasename, matrix, accession):
    ## log json = assert positions are correct?
    psite = {site: ratio for site, ratio in matrix.iloc[7].items() if ratio!=0}
    passertion = [duelKinase.loc[accession, 'STYD']]
    ttal = np.fromiter(psite.values(), dtype=float).sum()
    for ps, r in psite.items():
        duelKinase.loc[accession, f'KS_{ps}'] = 100*(r/ttal) ## if ratio is greater than 1 std above mean.
    duelKinase.loc[accession, [x for x in duelKinase.columns if 'KS' in x]] = duelKinase.loc[accession, \
                                                                            [x for x in duelKinase.columns\
                                                                            if 'KS' in x]].fillna(False)


def sanityCheck(kinasedata, pos=0.75):
    posk = kinasedata>= pos
    psiter = {p: 0  for p in ['S','T', 'Y']}
    posi = kinasedata[posk]
    for s, _ in psiter.items():
        posmask = posi.index.str[7].str.lower().str.contains(s.lower())
        pmean = posi[posmask].values.mean()
        psiter[s] = pmean*posmask.sum()
    print('finished!', kinasedata.name)
    logger.info(f'{kinasedata.name}_{pos}p', extra=psiter)
    return


def make_logo(kinasename, matrix, pos, accession):
    set_max = max(matrix.where(matrix >0).sum(axis=1))
    plt.figure(figsize=(42,30))
    COLORSCHEME = 'charge'
    STACKORDER='big_on_top'
    matrix = matrix.apply(lambda x: x.astype(float).fillna(0))
    logo = lm.Logo(matrix, color_scheme = COLORSCHEME,# font_name= 'Arial Rounded MT Bold', 
                        shade_below=0.3,
                        fade_below=0.2, 
                        flip_below=False,
                        # allow_nan=True,
                stack_order=STACKORDER)
    plt.gca().set_facecolor('white')
    plt.gcf().set_facecolor('white')
    plt.gca().set_xlabel('Position', color='black')
    plt.gca().set_ylabel('Relative Entropy', color='black') #
    plt.gcf().suptitle(f'{ORGANISM.capitalize()} kinase {kinasename} Substrate Motif', color='black', fontsize=18, y = 1) ## name to be workshopped
    plt.gca().spines['left'].set_color('black')
    plt.ylim(top=set_max)
    # plt.gca().set_title(f'{accession}',y=0.8 )
    plt.tick_params(axis='both', colors='black')  # Set the tick color
    plt.xticks(color='black')
    plt.yticks(color='black')
    filesave = f'../data/relative_entropy/{ORGANISM}_{str(pos).replace(".", "")}/{kinasename}.png'
    os.makedirs(os.path.dirname(filesave), exist_ok=True)
    plt.savefig(filesave,
            dpi=300, bbox_inches='tight', facecolor='white')
    return f'saved to {filesave}'


def get_probability(matrix):
    frequency = matrix.apply(lambda pos: pos.apply(lambda v: v/pos.sum()), axis=1)
    return frequency


def get_bits(matrix):
    norm = matrix * np.log2(len(matrix.columns) * matrix)
    norm = norm.apply(lambda x : x.astype(float).fillna(0))
    return norm


def scale_phosphosite(matrix): # parsing in KL divergence output
    masked = np.ones(matrix.shape[0], dtype=bool)
    masked[7]  = False  
    mx = (matrix[matrix[masked]>0].sum(axis=1)).max()
    scalar = mx/matrix.iloc[7].sum()
    matrix.iloc[7] = matrix.iloc[7] *scalar
    return matrix 


def kl_divergence(kin_dist, bkgrd_dist):
    significant_probabilities = kin_dist * np.log2(kin_dist/ bkgrd_dist)
    return significant_probabilities


def analyze_eachKinase(kinase, kinasedata, accession, bkgrd, pos=0.75):
    print('evaluating kinase: ', kinase, file=sys.stdout)
    symbols = ['A', 'R', 'N', 'D', 'C', 'E', 'Q', 'G', 'H', 'I', 'L', 'K', 'M', 'F', 'P', 'S', 'T', 'W', 'Y', 'V']
    count = pd.DataFrame(0, columns=symbols, index=range(-7, 8))
    posk = kinasedata[kinasedata>=pos]
    os.makedirs(f'../substrates/{ORGANISM}_{str(pos).replace(".", "")}', exist_ok=True)
    with open(f'../substrates/{ORGANISM}_{str(pos).replace(".", "")}/{kinase}.txt', 'wt') as f:
        for x in posk.index:
            print(f'>{x}', file = f)
            print(f'{x}', file=f)
    valid = posk[posk.index.str[7].str.upper().str.contains(r'S|T|Y')]
    logging.basicConfig(filename=f'{pos}{ORGANISM}.log', level=logging.INFO, format='%(message)s')
    logger.info(f'{kinase}:\tpos_substrates-{posk.shape[0]}')    
    for sub, pred in valid.items():
        for i, aa in enumerate(sub):
            if aa in ['_', 'X']:
                continue
            count.loc[i-7, aa.upper()] += 1*pred ## account for val
    count = count.apply(lambda x: x.astype(float).fillna(0))
    obs_frequency = get_probability(count) # p(i)
    bkgrd_frequency = make_bkgrd_matrix(bkgrd) # q(i)
    ## kl divergence == sum( p(k) * log2(p(k)/q(k)))  , individual is wihtout summation
    klDivergence = kl_divergence(obs_frequency, bkgrd_frequency)
    ## skip for position7
    klDivergence.iloc[7] = obs_frequency.iloc[7]
    scalednp = scale_phosphosite(klDivergence)
    print(make_logo(kinase, scalednp, pos, accession))
    if ORGANISM.lower() == 'human':
        assertPhosphosite(kinase, scalednp, accession)
    sanityCheck(kinasedata)
    return f'made logo for {kinase}'


global ORGANISM
ORGANISM = 'human'
taxID = {'human': 9606, 'mouse': 10090}
infile = f'../data/KS_genome_preds/cleaned_{ORGANISM.lower()}_kinase_substrate_predictions.csv'
data = pd.read_csv(infile, index_col=0)
uniprot2gene = IDmapping_main(data.columns.to_list(), 'UniProtKB_AC-ID', 'Gene_Name')
data = data.rename(columns={x: y for x, y in uniprot2gene.items() if y is not None})
global duelKinase
duelKinase = pd.read_excel(f'../data/Logos_STYE_KinaseNames_finalkeyV2_dualkinases.xlsx', 
                        sheet_name=0, index_col=0)
symbols = ['A', 'R', 'N', 'D', 'C', 'E', 'Q', 'G', 'H', 'I', 'L', 'K', 'M', 'F', 'P', 'S', 'T', 'W', 'Y', 'V']
## get background distributions
if os.path.exists(f'../data/{ORGANISM.lower()}_bkgrdDistrbutions.json'):
    allDistributions = json.load(open(f'../data/{ORGANISM.lower()}_bkgrdDistrbutions.json'))
else:
    print(f'creating background for {ORGANISM}, in file ../data/{ORGANISM.lower()}_bkgrdDistrbutions.json')
    allDistributions= {}
    substrates = data.index[data.index.str[7].str.upper().isin(['S','T','Y'])]
    with open(f'../data/{ORGANISM.lower()}_bkgrdDistrbutions.json', 'w') as writer:
        for ps in range(15):
            allDistributions[str(ps)] = get_background_distribution(substrates, position=ps)
        allDistributions['all']  =  get_background_distribution(substrates, position=None)
        json.dump(allDistributions, writer, indent=4)

logger           = logging.getLogger()
logger.setLevel(logging.INFO)
logHandler       = logging.FileHandler(f'{ORGANISM}.log')
formatter        = jsonlogger.JsonFormatter()
logHandler.setFormatter(formatter)
logger.addHandler(logHandler)

for i, col in tqdm(enumerate(data.columns), total = len(data.columns)):
    for v, x in enumerate(pd.DataFrame(data[col]).columns):
        if pd.DataFrame(data[col]).iloc[:, v].name not in ['Substrate', 'GENE', 'ACC_ID', 'MOD_RSD']: 
            curracc = [ac for ac, g in uniprot2gene.items() if g == x][v]
            kinasedata = pd.DataFrame(data[col]).iloc[:, v]
            out = analyze_eachKinase(kinasedata.name, kinasedata, curracc, allDistributions)

if ORGANISM.lower() == 'human':
    duelKinase.to_excel('../data/Logos_STYE_KinaseNames_finalkeyV2_dualkinases.xlsx')





from . import utils 
from . import models 

from .kolossus import run_kolossus
from .kolossus import kolossus
from .kolossus import set_batch_size


__version__ = '0.0.3'
__all__ = ['run_kolossus', 'kolossus']




import pickle 
import pandas as pd 


def main():
    fname = '../data/clinvar_to_protein_accessions_of_interest_with_seqs.pkl'
    df = get_seq_dataframe(fname)

    fname_out = '../data/clinvar_protein_sequences.csv'
    df.to_csv(fname_out)


def get_seq_dataframe(fname):
    with open(fname, 'rb') as f:
        results = pickle.load(f)
    
    data = []
    cols = ['accession', 'seq_accession', 'change', 'mane_select', 'original_seq', 'mutated_seq']
    for acc, dlist in results.items():
        for d in dlist:
            data.append(tuple([acc] + [d.get(c) for c in cols[1:]]))
    
    return pd.DataFrame(data, columns=cols)
            
        

if __name__ == '__main__':
    main()




# for reading and parsing data
from .read_data import convert_data_seqs_to_embeddings

from .read_data import get_model_input_from_sequences
from .read_data import get_model_input_from_embeddings

from .read_data import read_pairs
from .read_data import read_sequences
from .read_data import read_embeddings

from .seq_to_embedding import get_embeddings

# for batching operations
from . import batch_utils 

__all__ = []



#### misc operations on lists 

import math 


def batch_list(A, n):
    """
    Splits the list A into batches of size n. 
    """
    nbatches = math.ceil(len(A) / n)
    out = [None for _ in range(nbatches)]
    for i, j in enumerate(range(0, len(A), n)):
        out[i] = A[j:j+n]
    return out


def order_cols(A, f, *args, reverse=False, **kwargs):
    """
    Orders columns of a 2D-array A according to functional f. 

    (Assumes A is a numpy array or something similar like a tensor).
    """
    keys = list(range(len(A[0])))
    keys_sorted = sorted(keys, key=lambda i: f(A[:, i], *args, **kwargs), reverse=reverse)
    return keys_sorted, A[:, keys_sorted]





# general python imports 
import sys 
from functools import reduce

# for parsing input
from io import StringIO

from .utils import get_model_input_from_sequences
from .utils import get_model_input_from_embeddings

from .utils import read_sequences
from .utils import read_pairs 
from .utils import read_embeddings

from .utils import get_embeddings

from .utils import convert_data_seqs_to_embeddings

# for model 
from .models import load_model
import torch 

# for batching pairs 
from .utils.batch_utils import BatchComp

BATCH_SIZE = 10000
BATCH_HANDLER = BatchComp(BATCH_SIZE)

# for the conditional returns
from .utils.conditional_returns import DISCARD_COMPONENTS
LEN_T = 4
DISCARD = DISCARD_COMPONENTS(LEN_T, True)

# for warnings 
from .utils.warnings import warn

# load model for running 
MODEL = None


# FLAG 274: remove testing parameter for final version
@warn(274)
def run_kolossus(pairs, seqs=None, embeds=None, dtype=torch.float32, device='cpu', testing=False, 
                 return_projections=False):
    # 2 things to be done if seqs are input: 
    #     1. convert sequence to embeddings
    #     2. convert elements of pairs from triplets to doublets  
    # task 1 and task 2 done simultaneously in function `convert_data_seqs_to_embeddings`
    if seqs is not None:
        assert isinstance(seqs, dict)
        pairs, embeds = convert_data_seqs_to_embeddings(seqs, pairs, device, testing)
    else:
        assert embeds is not None 

    # don't store projections if not desired 
    if not return_projections:
        DISCARD.add_discard_index(-1, -2)

    # will operate on pairs in batches 
    result = reduce(lambda x, y: x + y, _run_kolossus(pairs, embeds, dtype, device))

    # return either just distances or distances and projections
    out = kolossus_output(result, return_projections)
    return out

    

def kolossus(fpairs, fseqs=None, fembeds=None, dtype=torch.float32, device='cpu', 
             return_projections=False):
    if fseqs is not None:
        seqs = read_sequences(fseqs, to_dict=True)
        pairs = read_pairs(fpairs, '\t', includes_window=True)
        return run_kolossus(pairs, seqs=seqs, dtype=dtype, device=device, 
                            return_projections=return_projections)
    else:
        assert fembeds is not None 
        # read model embeddings
        embeds = read_embeddings(fembeds)
        # assume each pair embedding is for the appropriate substrate window
        pairs = read_pairs(fpairs, '\t', includes_window=False)
        return run_kolossus(pairs, embeds=embeds, dtype=dtype, device=device,
                            return_projections=return_projections)


# where the magic happens
@BATCH_HANDLER
@DISCARD
def _run_kolossus(pairs, embeds, dtype, device):
    global MODEL
    if MODEL is None: 
        MODEL = load_model()

    X = get_model_input_from_embeddings(embeds, pairs, dtype, device)

    MODEL.eval()
    MODEL.to(device)
    with torch.no_grad():
        Y, kinase_embeds, site_embeds = MODEL(X)

    model_out = tuple(map(lambda x: x.detach().cpu().numpy().tolist(),
                        (Y, kinase_embeds, site_embeds)))

    return tuple(zip(pairs, *model_out))


def kolossus_output(result, return_projection):
    if return_projection:
        # get distances for pairs 
        distances = dict(map(lambda t: (t[0], t[1]), result))
        embeddings = {}
        for ((kid, sid), _, kembed, sembed) in result:
            embeddings[kid] = kembed
            embeddings[sid] = sembed
        return distances, embeddings
    else:
        return dict(result)


def set_batch_size(n):
    global BATCH_SIZE
    BATCH_SIZE = n
    try:
        BATCH_HANDLER.set_block_size(BATCH_SIZE)
    except AssertionError:
        print(f"Error: invalid block size passed in: {n}.", file=sys.stderr)
        print("Block size must be an integer of size at least 2", file=sys.stderr)
        sys.exit(1)
    




import os
import sys
import pathlib

import tempfile
import h5py
from io import StringIO

import numpy as np

import torch 
import esm 
from esm import FastaBatchedDataset, pretrained
from transformers import AutoTokenizer, AutoModel

# for handling padding characters 
from .read_data import PADDING_CHAR

# warnings for testing 
from .warnings import warn

# for writing fasta 
from .write_fasta import write_fasta


# FLAG 274: for Kanchan to test the code and make sure it's functional
# NOTE: remove testing in final version
# seq_list is a list of tuples (seq_id, seq)
@warn(274)
def get_embeddings(seq_list, device, testing=False):
    if testing:
        return _get_embeddings_testing(seq_list)
    
    # print("ERROR: UNTESTED CODE: KANCHAN PLEASE TEST THIS OUT IN FILE", __file__, "!!!", file=sys.stderr)
    # sys.exit(1)
    
    # again, for Kanchan to test this out and debug
    # Load ESM-2 model
    model, alphabet = load_model()
    batch_converter = alphabet.get_batch_converter()
    model.to(device)
    model.eval()  # disables dropout for deterministic results

    # Prepare data (first 2 sequences from ESMStructuralSplitDataset superfamily / 4)
    data = [(a, s.strip(PADDING_CHAR)) for a, s in seq_list]
    batch_labels, batch_strs, batch_tokens = batch_converter(data)
    batch_lens = (batch_tokens != alphabet.padding_idx).sum(1)

    # Extract per-residue representations (on CPU)
    with torch.no_grad():
        results = model(batch_tokens, repr_layers=[48], return_contacts=True)
    token_representations = results["representations"][48]

    # Generate per-sequence representations via averaging
    # NOTE: token 0 is always a beginning-of-sequence token, so the first residue is token 1.
    sequence_representations = []
    for i, ((seq_id, _), tokens_len) in enumerate(zip(seq_list, batch_lens)):
        seq_rep = token_representations[i, 1 : tokens_len - 1].mean(0)
        sequence_representations.append((seq_id, seq_rep))

    return sequence_representations
    

def _get_embeddings_testing(seq_list):
    d = {'A': 0, 'C': 1, 'G': 2, 'T': 3, PADDING_CHAR: 4}
    out = [None for _ in range(len(seq_list))]
    dim = 5120
    for i, (seq_id, s) in enumerate(seq_list):
        x = torch.Tensor([d[c] for c in s] + [-100 for _ in range(dim-len(s))])
        out[i] = (seq_id, x)
    
    return out


# alternative function for getting embeddings
def extract_embeddings(seq_list, device, model_name='esm2_t48_15B_UR50D', output_dir=None, 
                       tokens_per_batch=4096, seq_length=1022,repr_layers=[48]):
    # handling defaults
    if output_dir is None:
        output_dir = pathlib.Path('.')
    else:
        if isinstance(output_dir, str):
            assert os.path.isdir(output_dir)
            output_dir = pathlib.Path(output_dir)
        assert isinstance(output_dir, pathlib.Path)

    print('-' * 40)
    print("Loading esm model")
    model, alphabet = pretrained.load_model_and_alphabet(model_name)
    print("Esm model done loading")
    model.eval()

    if torch.cuda.is_available():
        model.to(device)
    else:
        model.to('cpu')

    with tempfile.TemporaryDirectory() as tempdir:
        # make fasta file and delete disk storage of sequences 
        fasta_file = f'{tempdir}/seqs.fasta'
        write_fasta(list(map(lambda t: (t[0], t[1].strip(PADDING_CHAR)), seq_list)), fasta_file)
        nseqs = len(seq_list)
        del seq_list
        
        dataset = FastaBatchedDataset.from_file(pathlib.Path(fasta_file))
        batches = dataset.get_batch_indices(tokens_per_batch, extra_toks_per_seq=1)
    
        data_loader = torch.utils.data.DataLoader(
            dataset,
            collate_fn=alphabet.get_batch_converter(seq_length),
            batch_sampler=batches)
    
        output_dir.mkdir(parents=True, exist_ok=True)
        
        with torch.no_grad(): 
            for batch_idx, (labels, strs, toks) in enumerate(data_loader):

                print(f'Processing batch {batch_idx + 1} of {len(batches)}')

                if torch.cuda.is_available():
                    toks = toks.to(device=device, non_blocking=True)

                out = model(toks, repr_layers=repr_layers, return_contacts=False)
                logits = out["logits"].to(device="cpu")
                representations = {layer: t.to(device="cpu") for layer, t in out["representations"].items()}

                for i, label in enumerate(labels):
                    entry_id = label.split()[0]

                    filename = output_dir / f"{tempdir}/{entry_id}.pt"
                    truncate_len = min(seq_length, len(strs[i]))

                    result = {"entry_id": entry_id}
                    result["mean_representations"] = {
                            layer: t[i, 1 : truncate_len + 1].mean(0).clone()
                            for layer, t in representations.items() }
    
                    torch.save(result, filename)

        # get embeddings that are saved 
        print("extracting mean representations of sequences")
        fnames = list(filter(lambda s: s.endswith('.pt'), os.listdir(tempdir)))
        out = {}
        for fname in fnames:
            fname = os.path.join(tempdir, fname)
            sid = os.path.split(fname)[-1].split('.')[0]
            sx = torch.load(fname)['mean_representations'][48]
            out[sid] = sx
        print(f"done converting {len(out)} / {nseqs} sequences to embeddings")
        print('-'*40)

    return out


def extract_embeddings_from_fasta(fasta_file, output_file, device, model_name, 
                       tokens_per_batch=4096, seq_length=1022,repr_layers=[48],layer_to_use=48):
    # assert valid output and input files 
    assert os.path.isfile(fasta_file)
    assert os.path.isdir(os.path.join(*os.path.split(output_file)[:-1]))
    if not output_file.endswith('.h5'):
        output_file += '.h5'

    print('-' * 40)
    print("Loading esm model")
    model, alphabet = pretrained.load_model_and_alphabet(model_name)
    print("Esm model done loading")
    model.eval()

    if torch.cuda.is_available():
        model.to(device)
    else:
        model.to('cpu')

    dataset = FastaBatchedDataset.from_file(pathlib.Path(fasta_file))
    batches = dataset.get_batch_indices(tokens_per_batch, extra_toks_per_seq=1)

    data_loader = torch.utils.data.DataLoader(
        dataset,
        collate_fn=alphabet.get_batch_converter(seq_length),
        batch_sampler=batches)

    with torch.no_grad(): 
        for batch_idx, (labels, strs, toks) in enumerate(data_loader):
    
            print(f'Processing batch {batch_idx + 1} of {len(batches)}')
    
            if torch.cuda.is_available():
                toks = toks.to(device=device, non_blocking=True)
    
            out = model(toks, repr_layers=repr_layers, return_contacts=False)
            logits = out["logits"].to(device="cpu")
            representations = {layer: t.to(device="cpu") for layer, t in out["representations"].items()}
    
            with h5py.File(output_file, 'a') as fout:
                for i, label in enumerate(labels):
                    entry_id = label
                    truncate_len = min(seq_length, len(strs[i]))
    
                    key = entry_id
                    mean_representations = {
                            layer: t[i, 1 : truncate_len + 1].mean(0).clone()
                            for layer, t in representations.items() }
    
                    fout[key] = mean_representations[layer_to_use]

    print("saved all sequences to", output_file)



# in case we need to modify later
def load_model():
    return esm.pretrained.esm2_t48_15B_UR50D()





from .utils.seq_to_embedding import extract_embeddings_from_fasta

import argparse
import sys 
import pathlib
import os 


def main():
    args = parse_args()
    
    fasta_file = args.i
    model_name = args.model
    device = args.device
    output_file = args.o

    extract_embeddings_from_fasta(fasta_file, output_file, device, model_name)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', type=str, required=True, help='name of input fasta file')
    parser.add_argument('--model', type=str, default='esm2_t48_15B_UR50D', help='name of model to extract embeddings')
    parser.add_argument('--device', type=str, default='cpu', help='cpu or gpu device to use')
    parser.add_argument('-o', type=str, required=True, help='name of output .h5 file')
    args = parser.parse_args()
    return args 


if __name__ == '__main__':
    main()




from . import stringops
from .decorators import batch_comp_on_list
from .decorators import delay

from io import StringIO
from functools import reduce

from Bio import SeqIO
from Bio.Seq import Seq 
from Bio.SeqRecord import SeqRecord 

from Bio.Entrez import efetch
from Bio import Entrez

from urllib.error import HTTPError


def main():
    print(Entrez.email)
    seqs = {'seq1': 'aaccggtt',
            'seq2': 'aaccggt',
            'seq3': 'aaccggttt'}
    n = 4
    outfile = 'test.fasta'

    write_fasta(seqs, outfile, n)

    seqs = [('seq1', 'aaccggtt'),
            ('seq2', 'aaccggt'),
            ('seq3', 'aaccggttt')]
    n = 4
    outfile = 'test2.fasta'

    write_fasta(seqs, outfile, n)


def write_fasta(seqs, outfile, n=80):
    """
    Write sequences to fasta file. Note that we assume the order of the sequences
    is unimportant if the input passed in is a dictionary. 

    Input: 
        seqs (dict[str] -> str) OR (list[(str, str)]): a dictionary or a list of tuples (fasta header, sequence)
        outfile (str): path for output file
    """
    if isinstance(seqs, dict):
        seqs = seqs.items()
    with open(outfile, 'wt') as fout:
        for header, seq in seqs:
            print(f'>{header}', file=fout)
            print('\n'.join(stringops.chop(seq, n)), file=fout)


def fetch_seqs(db, id, rettype='fasta', retmode='text', parser=None):
    # must switch order of input in backend to accomodate batch decorator
    batches = reduce(lambda x, y: x + y, _fetch_seqs(id, db, rettype, retmode, parser))
    return {name: seq for name, seq in batches}


@batch_comp_on_list(block_size=150)
@delay(limit=0.35, time_delay=0.5)
def _fetch_seqs(id, db, rettype, retmode, parser):
    if not parser:
        if rettype == 'fasta' and retmode == 'text':
            parser = parse_entrez_fasta_output
    assert parser is not None
    try:
        seqs = efetch(db=db, id=id, rettype=rettype, retmode=retmode).read()
    except HTTPError:
        return None
    return parser(seqs)


def parse_entrez_fasta_output(s):
    ssss = '\n'.join(list(filter(lambda sss: len(sss) > 0, [ss.strip() for ss in s.split('\n')])))
    records = SeqIO.parse(StringIO(ssss), 'fasta')
    return [(r.id, str(r.seq)) for r in records]


if __name__ == '__main__': 
    main()



import re
import pandas as pd
import numpy as np

import argparse

from Bio import SeqIO


# argument parser defaults 
MASTER_KINASE_LIST = '../data/STYE_KinaseNames_finalkeyV2.csv'


def main():
    # parse arguments
    args = parse_args()
    
    # load variant data 
    fname = args.i
    if args.o is None:
        fname_out = fname[:-4] + '_filtered.txt'
    else:
        fname_out = args.o
    df = pd.read_csv(fname, delimiter='\t') 
    
    print("shape of unfiltered data:", df.shape)
    
    # load kinase names 
    kinase_names = pd.read_csv('../data/STYE_KinaseNames_finalkeyV2.csv')

    # drop proteins where the ProteinName column is undefined
    kinase_names = kinase_names.dropna(axis=0, subset=['ProteinName'])

    # get kinase sequences 
    records = SeqIO.parse('../../../alt_splicing/rawdata/uniprot_sprot.fasta', 'fasta')
    query = r'^.+OS=Homo sapiens.+GN=(\S+).+$'
    records = list(filter(lambda rec: re.search(query, rec.description) is not None, records))
    records = dict(map(lambda rec: (re.search(query, rec.description).group(1), str(rec.seq)), 
                       records))

    kinase_names['Kinase_seq'] = kinase_names['ProteinName'].apply(lambda name: records[name])
    
    # filter out non-valid kinase entries  
    filtered_df = filter_kinases(df, kinase_names)
    
    # print shape of final data 
    print("filtered dataframe shape:", filtered_df.shape)
    
    # write to file 
    filtered_df.to_csv(fname_out)
    print("wrote filtered dataframe to file", fname_out)


def get_protein_change_from_name(s, query):
    search = re.search(query, s)
    if search:
        return search.group(2)
    else:
        return np.nan
        

def filter_kinases(variant_data, kinase_info):
    query_protein_name = r'^NM_[0-9.]+\((.+)\):c\.\S+\s\(p.([A-Za-z]{3}[0-9]+[A-Za-z]{3})\)'
    query_gene_names = r'^NM_[0-9.]+\((.+)\).+$'
    
    kinase_set = set(kinase_info['ProteinName'].tolist())
    
    # filtering based on kinase_names 
    is_kinase = variant_data['Name'].apply(lambda s: (re.search(query_gene_names, s) is not None) and (re.search(query_gene_names, s).group(1) in kinase_set))
    
    variant_data = variant_data[is_kinase]
    
    return variant_data 


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', type=str, required=True, help='name of input variant file')
    parser.add_argument('-o', type=str, required=False, default=None, help='name of output (filtered) variant file')
    parser.add_argument('--kinase-list', type=str, required=False, default=MASTER_KINASE_LIST, help='name of master kinase list file')
    args = parser.parse_args()
    return args 


if __name__ == '__main__': 
    main()



from .kolossus import BATCH_SIZE
from .kolossus import set_batch_size
from .kolossus import kolossus

import argparse
import sys 
import h5py


def main():
    args = parse_args()

    if args['batch_size']:
        print(f"setting batch size to {args['batch_size']}...", end=' ')
        set_batch_size(args['batch_size'])
        print("done!")

    print("running model...", end=' ')
    results = kolossus(**args['kolossus'])
    print("done!")

    output_results(results, **args['output'])


def output_results(results, dists_fname_out, projections_fname_out):
    if projections_fname_out:
        results, projections = results
        print("writing embeddings to:", projections_fname_out)
        with h5py.File(projections_fname_out, 'w') as fout:
            for _id, e in projections.items():
                fout[_id] = e

    print("writing predictions to", dists_fname_out)
    with open(dists_fname_out, 'wt') as fout:
        print('#kinase\tsubstrate\tpredicted_prob', file=fout)
        for pair, prob in results.items():
            pair = tuple(map(str, pair))
            prob = str(prob)
            print('\t'.join((*pair, prob)), file=fout)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--pairs', type=str, required=True, 
                        help="format: <kinase_id>\t<substrate_id>\t<substrate_phosphorylation_site>")
    parser.add_argument('-s', '--seqs', type=str, default=None, 
                        help='fasta formatted file of sequences, either seqs or embeddings must be provided')
    parser.add_argument('-e', '--embeddings', type=str, default=None, 
                        help='h5 file of sequence embeddings, either seqs or embeddings must be provided')
    parser.add_argument('--dtype', type=str, default='', 
                        help='data type of sequence embeddings (usually float32)')
    parser.add_argument('-r', '--projections', type=str, default='', 
                        help='name of .h5 files for kolossus projections')
    parser.add_argument('-d', '--device', type=str, default='cpu', 
                        help='default device on which to run model')
    parser.add_argument('--batch_size', type=int, default=None, 
                        help='Number of pairs at a time on which to run model')
    parser.add_argument('-o', '--output', type=str, required=True, 
                        help='desired file path for output')

    args = parser.parse_args()

    real_args = {'kolossus': {'fpairs': args.pairs,
                         'fseqs': args.seqs,
                         'fembeds': args.embeddings,
                         'dtype': args.dtype,
                         'device': int(args.device) if args.device.isdigit() else args.device,
                         'return_projections': len(args.projections) > 0},
                 'batch_size': args.batch_size,
                 'output': {'dists_fname_out': args.output,
                            'projections_fname_out': args.projections if len(args.projections) > 0 else None}}
    
    return real_args    


if __name__ == '__main__':
    main()




import sys
from collections import defaultdict 

# for seqs
from .parse_fasta import parse_fasta 
from .seqops import extract_window

# FLAG 274: consult with Kanchan as far as how to pad 
PADDING_CHAR = '#'
WINDOW_SIZE = 15

# for embeddings
import h5py
import torch
# from .seq_to_embedding import get_embeddings
from .seq_to_embedding import extract_embeddings

# warnings for testing 
from .warnings import warn


# FLAG 274: for kanchan: how to handle padding for substrate sites near start or end of sequence 
# assume pairs are in form (kinase_sequence_id, substrate_sequence_id, substrate_phosphorylation_site), tab or comma separated
@warn(274)
def read_pairs(fname, delimiter='\t', includes_window=True):
    pairs = []
    with open(fname, 'rt') as f:
        for line in f:
            line = line.strip()
            if line.startswith('#'):
                continue
            spl = line.split(delimiter)
            spl = [s.strip() for s in spl]
            if includes_window:
                tup = (spl[0], spl[1], int(spl[2]))
            else:
                tup = (spl[0], spl[1])
            pairs.append(tup)
    return pairs 


def read_sequences(fname, to_dict=False):
    return parse_fasta(fname, to_dict)


# assumes an h5py file of the embeddings
# returns dictionary of the embeddings to numpy arrays
# note that in this case we assume each (substrate, phospho_site) is annotated by record keys 
def read_embeddings(fname, dtype=torch.float32, device='cpu'):
    embeddings = {}
    with h5py.File(fname, 'r') as f:
        for key in f.keys():
            embedd = torch.from_numpy(f[key][:])
            embedd = embedd.to(dtype)
            embedd = embedd.to(device)
            embeddings[key] = embedd
        
    return embeddings 


# FLAG 274: remove testing parameter for final version
@warn(274)
def build_model_input_from_sequences(ffasta, fpairs, delimiter='\t', dtype=torch.float32, device='cpu', testing=True):
    sequences = read_sequences(ffasta, to_dict=True)
    pairs = read_pairs(fpairs, delimiter, includes_window=True)
    return get_model_input_from_sequences(sequences, pairs, dtype=dtype, device=device, testing=True)


# FLAG 274: remove testing parameter for final version
@warn(274)
def convert_data_seqs_to_embeddings(sequences, pairs, device, testing=True):
    # get (substrate, site) from pairs 
    subs_to_site = defaultdict(set)
    for _, subseq_id, site in pairs:
        subs_to_site[subseq_id].add(site)

    # start the list off with the full kinase sequences 
    kinase_ids = set([t[0] for t in pairs])
    # note that some in some cases (e.g., panel), kinase embeddings may already be provided
    seqlist = [(k, sequences[k]) for k in kinase_ids if k in sequences]
    
    # now extract the corresponding 15-mers 
    for subseq_id, siteset in subs_to_site.items():
        for site in siteset:
            kmer_id = make_kmer_id(subseq_id, site) 
            subseq = sequences[subseq_id]
            window = extract_window(subseq, site, WINDOW_SIZE, PADDING_CHAR)
            seqlist.append((kmer_id, window))

    # update pairs
    pairs = list(map(lambda t: (t[0], make_kmer_id(t[1], t[2])), pairs))

    # extract the embeddings for all sequences 
    # seq_embeddings = {k: v for k, v in get_embeddings(seqlist, device)}
    seq_embeddings = extract_embeddings(seqlist, device)

    return pairs, seq_embeddings


# FLAG 274: remove testing parameter for final version
# we probably won't use this function
@warn(274)
def get_model_input_from_sequences(sequences, pairs, dtype=torch.float32, device='cpu', testing=True):
    pairs, seq_embeddings = convert_data_seqs_to_embeddings(sequences, pairs, dtype=torch.float32, device='cpu', testing=True)
    
    # now we can simply return the pair of output embeddings 
    return get_model_input_from_embeddings(seq_embeddings, pairs, dtype=dtype, device=device)


def build_model_input_from_embeddings(fembed, fpairs, delimiter='\t', dtype=torch.float32, device='cpu'):
    # read model embeddings
    seq_embeddings = read_embeddings(fembed)
    # assume each pair embedding is for the appropriate substrate window
    pairs = read_pairs(fpairs, delimiter, includes_window=False)
    return get_model_input_from_embeddings(seq_embeddings, pairs, dtype=torch.float32, device='cpu')


def get_model_input_from_embeddings(seq_embeddings, pairs, dtype=torch.float32, device='cpu'):
    # make the tensors 
    out = [None for _ in range(len(pairs))]
    for i, (kinase_id, substrate_id) in enumerate(pairs):
        x1 = seq_embeddings[kinase_id]
        x2 = seq_embeddings[substrate_id]

        # FLAG 274: substrate goes into first projector
        x = torch.concat([x2, x1])

        # unsqueeze to allow observations concatenated together 
        out[i] = x.unsqueeze(0)

    # concatenate all observations into single 2-D tensor and return 
    return torch.concat(out).to(device)

    
def make_kmer_id(seqid, site):
    return f'{seqid}:psite={site}'
    




import gzip
import pickle

import re

import pandas as pd 
import numpy as np

# for custom functions 
from clinvar_xml_parsing import get_variation_archive_info


def main():
    # note: I already checked, no accession overlap
    fnames = {'benign': '../data/clinvar_result_msmuts_benign_0930_filtered.txt',
              'pathogenic': '../data/clinvar_result_msmuts_pathogenic_09_24_filtered.txt'}
    
    accs = {key: set(pd.read_csv(fnames[key])['Accession'].tolist()) for key in fnames}
    
    # these are the accessions of interest 
    accessions_of_interest = accs['benign'].union(accs['pathogenic'])
    
    # query for start and end of variation archive 
    query_start = r'<VariationArchive.*>'
    query_end = r'</VariationArchive.*>'

    A = {}
    block_ct = 0    # target: 508,908
    with gzip.open('/hpc/group/singhlab/rawdata/ClinVarVCVRelease_00-latest.xml.gz', 'rt') as f:
#         ct = 0
        while True:
            line = f.readline()

            if not line: 
                break 

            # look for beginning of variation archive record 
            if re.search(query_start, line):
                accession, info = get_variation_archive_info(line, f, query_end)

                
                block_ct += 1
                # only save records for archives of interest 
                if accession not in accessions_of_interest:
                    continue
                    
                if accession not in A:
                    A[accession] = info
                else:
                    A[accession].append(info)
                
                print("wrote accession", accession, "to A,", len(A), "/", len(accessions_of_interest), "mapped so far")

#             ct += 1
#             if ct > 2000:
#                 break 

    print("parsed", block_ct, "blocks to obtain", len(A), "records")
    with open("../data/clinvar_to_protein_accessions_of_interest.pkl", 'wb') as f:
        pickle.dump(A, f)
    

if __name__ == '__main__': 
    main()




import torch
import numpy as np

import h5py

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

AMINO_ACIDS = ['A', 'R', 'N', 'D', 'C', 'Q', 'E', 'G', 'H', 'I', 'L', 'K', 'M', 'F', 'P', 'S', 'T', 'W', 'Y', 'V']


def main():
    n = 9
    fout_seqs = 'test.fasta'
    fout_pairs = 'pairs.txt'
    fout_embed = 'embed.h5'

    embedding_dim = 5120

    kinase_seqs = [(f'kin{i+1}', s) for i, s in enumerate(rseqs(n))]
    substrate_seqs = [(f'sub{i+1}', s) for i, s in enumerate(rseqs(n))]
    seqs = kinase_seqs + substrate_seqs
    tofasta(seqs, fout_seqs)

    pairs_out = set()
    for (a, _), (b, _) in zip(kinase_seqs, substrate_seqs):
        n = len(b)
        i = str(np.random.choice(n))
        pairs_out.add((a, b, i))
    with open(fout_pairs, 'wt') as fout:
        print('#lines starting with \'#\' are ignored', file=fout)
        print('#kinase_id\tsubstrate_id\tphosphorylation_site', file=fout)
        for pair in pairs_out:
            print('\t'.join(pair), file=fout)

    with h5py.File(fout_embed, 'w') as fout:
        for name, s in seqs:
            x = torch.Tensor([AMINO_ACIDS.index(c) for c in s] + [-100 for _ in range(embedding_dim-len(s))])
            fout[name] = x


def rseqs(n, maxlen=60, minlen=20):
    lens = (np.random.choice(range(minlen, maxlen), size=n, replace=True) + 1).tolist()
    seqs = [''.join(np.random.choice(AMINO_ACIDS, size=a, replace=True).tolist()) for a in lens]
    return seqs


def tofasta(seqs, fout, prefix=''):
    seqs = [SeqRecord(Seq(s), id=name, description='') for name, s in seqs]
    SeqIO.write(seqs, fout, 'fasta')


if __name__ == '__main__':
    main()




import re 

import pickle 
import argparse 

import os
import sys
import time

import functools

# string io operations 
from io import StringIO

# BioPython
from Bio import Entrez
from Bio.Entrez import efetch
from Bio import SeqIO

from Bio.Data.IUPACData import protein_letters_3to1

from urllib.error import HTTPError

Entrez.email = 'aditya.parekh@duke.edu'

# custom functions 
sys.path.append('/hpc/home/aap100/src')
from aprkh_utils.stringops import apply_subs
from aprkh_utils import delay 
from aprkh_utils import batch_list

# for parsing mutations 
# Note that there are three queries which have complex (non-simple) substitutions:
# ['VCV000003030', 'VCV000140897', 'VCV001701565']
SUB_QUERY = re.compile(r'^p.([A-Z][a-z]{2})(\d+)([A-Z][a-z]{2})$')
SUB_QUERY2 = re.compile(r'^p.([A-Z][a-z]{2}\d+(_([A-Z][a-z]{2})(\d+))+)delins(([A-Z][a-z]{2})+)$')

SIMPLE_QUERY = re.compile(r'^p.([A-Z][a-z]{2})(\d+)([A-Z][a-z]{2})$')
DELINS_QUERY = re.compile(r'^p.(\S+)delins(\S+)$')


def main():
    fname = '../data/clinvar_to_protein_accessions_of_interest.pkl'
    fname_out = '../data/clinvar_to_protein_accessions_of_interest_with_seqs.pkl'
    
    with open(fname, 'rb') as f:
        results_dict = pickle.load(f)
    
    # get sequence accession list 
    seq_acc_list = functools.reduce(lambda x, y: x + y, [[d['seq_accession'] for d in v] for v in results_dict.values()])
    print('number of sequence accessions:', len(seq_acc_list))
    # map seq accessions to sequences
    fname_seqs = '../data/fetch_wt_and_mut_seqs_sequences.pkl'  # for saving progress 
    seq_acc_to_seqs = {k: None for k in seq_acc_list}
    # load partial progress
    if os.path.isfile(fname_seqs):
        with open(fname_seqs, 'rb') as f:
            seq_acc_to_seqs.update(pickle.load(f))
    # remove keys we already have sequences for 
    seq_acc_list = [k for k, v in seq_acc_to_seqs.items() if v is None]
    seq_acc_list = batch_list(seq_acc_list, 150)  # batch by 150 to make retrieval more efficient
    for i, batch in enumerate(seq_acc_list):
        seqs_batch = get_wild_type(batch)
        seq_acc_to_seqs.update(seqs_batch)
        
        print("on batch", i+1, "/", len(seq_acc_list))
        if i % 5 == 0:
            with open(fname_seqs, 'wb') as fout:
                print("saving partial sequences dictionary to", fname_seqs)
                pickle.dump(seq_acc_to_seqs, fout)
    
    with open(fname_seqs, 'wb') as fout:
        print("saving full sequences dictionary to", fname_seqs)
        pickle.dump(seq_acc_to_seqs, fout)
    
    # update the results dictionary 
    for var_accession, missense_list in results_dict.items():
        for change_dict in missense_list:
            seq_acc = change_dict['seq_accession']
            change = change_dict['change']
            
            # update sequence
            seq = seq_acc_to_seqs[seq_acc]
            if not seq_acc_to_seqs[seq_acc]:
                continue
                
            change_dict['original_seq'] = seq
            
            try: 
                if re.search(SUB_QUERY2, change):
                    change_dict['mutated_seq'] = apply_multiple_mutation(seq, change)
                else:
                    assert re.search(SUB_QUERY, change)
                    change_dict['mutated_seq'] = apply_simple_mutation(seq, change)
            except KeyError:
                print("skipping variant", var_accession, "of sequence", seq_acc, "with change", change)
                
    # now save the results
    with open(fname_out, 'wb') as fout:
        print("saving final results to", fname_out)
        pickle.dump(results_dict, fout)


@delay(limit=0.35, time_delay=0.35)  # to avoid getting banned by NCBI
def get_wild_type(accession_list):
    try:
        s_fasta = efetch(db='protein', id=accession_list, rettype='fasta', retmode='text').read()
    except HTTPError as e:
        raise e
    return parse_efetch_output(s_fasta)


def parse_efetch_output(s_fasta):
    s_split = s_fasta.split('\n')
    s_split = list(filter(lambda s: len(s.strip()) > 0, s_split))
    s_fasta = '\n'.join(s_split)
    records = list(SeqIO.parse(StringIO(s_fasta), 'fasta'))
    return {r.id: str(r.seq) for r in records}


# def apply_mutation
def apply_simple_mutation(s, change):
    # we assume this is a valid change and the regex will match change string
    before, index, after = re.findall(SIMPLE_QUERY, change)[0]
    before = protein_letters_3to1[before]
    after = protein_letters_3to1[after]
    subs = ((before, after, int(index)),)
    return apply_subs(s, subs, zero_based=False)


def apply_multiple_mutation(s, change):
    # we assume this is a valid change and the regex will match change string
    before, after = re.findall(DELINS_QUERY, change)[0]
    
    # parse before 
    before_aa = []
    for aanum in before.split('_'):
        aa = protein_letters_3to1[aanum[:3]]
        num = aanum[3:]
        assert num.isdigit()
        before_aa.append((aa, int(num)))
    
    # parse after 
    after_aa = [protein_letters_3to1[after[i:i+3]] for i in range(0, len(after), 3)]
    
    # should have same number of amino acids deleted and inserted 
    if len(before_aa) != len(after_aa):
        return None 
    
    # otherwise apply substitutions 
    subs = [(aa, bb, ii) for (aa, ii), bb in zip(before_aa, after_aa)]
    return apply_subs(s, subs, zero_based=False)
    
    
    
    
    
    
if __name__ == '__main__': 
    main()




import numpy as np
import pandas as pd


df = pd.read_csv('human_kinase_substrate_predictions.csv', nrows=500)
df['seqid'] = [f'seq{i+1}' for i in range(len(df))]
df = pd.melt(df, ['Substrate', 'seqid'], var_name='kinase', value_name='prob')

import re 
query = r'^seq[0-9]+$'
#query = 'boogaloo'
def fix(s):
    return f'{s}:psite=7' if re.search(query, s) else s
d1 = dict(map(lambda t: ((t[0], fix(t[1])), t[2]), df[['kinase', 'seqid', 'prob']].values.tolist()))

with open('results.txt', 'rt') as f:
    lines = [line.strip().split() for line in f.readlines() if not line.startswith('#')]

d2 = dict(map(lambda t: ((t[0], t[1]), float(t[2])), lines))

a = []
for k in d2:
    a.append(abs(d1[k] - d2[k]) < 1e-4)

print(all(a))

def cos_sim(x, y):
    return np.dot(x, y) / (np.sqrt(x@x) * np.sqrt(y@y))

import h5py 
with h5py.File('results.h5', 'r') as f1, open('results.txt') as f2:
    pp = []
    cc = []
    for line in f2:
        if line.strip().startswith('#'):
            continue
        k1, k2, p = line.split()
        pp.append(float(p))
        cc.append(cos_sim(f1[k1][:], f1[k2][:]))
print(np.allclose(*list(map(np.array, [pp, cc]))))




import torch
from .Network import Network

# note: very hacky, change to host on some website in the future 
import os
import sys 


MODEL_FILE = 'new_ft_v3_humansty.pth'

if not os.path.isfile(os.path.join(os.path.split(__file__)[0], MODEL_FILE)):
    print(f"Error: please save file {MODEL_FILE} to", os.path.split(__file__)[0], 
        "sorry, very bad programming practice, will be fixed later", file=sys.stderr)
    sys.exit(1)


def load_model():
    path = os.path.split(__file__)[0]
    model_weights = os.path.join(path, MODEL_FILE)
    model = Network()
    model.load_state_dict(torch.load(model_weights))
    return model 




import functools
from functools import wraps 

import sys 
import inspect

# FLAGS to be concerned about 
FLAGS = {274: ('print_warning_message_flag_274', 'check_testing')}


def main():
    foo(False) 
    foo(True)
    foo(False)


# decorators to wrap function 
def warn(flag, warn_condition=None):

    def decorator(func):
    
        @wraps(func)
        def wrapper(*args, **kwargs):
            wfn, cfn = FLAGS.get(flag, ('all_good', 'always_true'))
            warn_func = globals().get(wfn)
            cond_func = globals().get(cfn)

            # figure out whether or not we have to warn
            if cond_func(func, *args, **kwargs):
                # get function info 
                name = func.__name__
                file = inspect.getfile(func)
                line_num = inspect.getsourcelines(func)[1]
                warn_func(flag, name, file, line_num)
                
            result = func(*args, **kwargs)
            
            return result
        return wrapper

    return decorator


@warn(274)
def foo(testing):
    print('foo')


# printing warning messages 
def all_good(flag, name, file, line_num, *args, **kwargs):
    err("WARNING: INVALID FLAG CALLED:", flag)
    err(f"SEE FUNCTION {name} IN FILE {file}, LINE {line_num}")
    

# special decorator called here 
def print_warning_message_flag_274(flag, name, file, line_num):
    err("WARNING: FLAG 274: OBTAINING JUNK SEQUENCES", end='! ')
    err(f"SEE FUNCTION {name} IN FILE {file}, LINE {line_num}")


# condition checkers
def always_true(func, *args, **kwargs):
    return True


def check_testing(func, *args, **kwargs):
    arg_values = inspect.getcallargs(func, *args, **kwargs)
    if arg_values.get('testing') == True:
        return True
    return False
    

def err(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

if __name__ == '__main__': 
    main()




from .kolossus import BATCH_SIZE
from .kolossus import set_batch_size
from .kolossus import run_kolossus

from .utils import read_sequences
from .utils import read_pairs 
from .utils import read_embeddings
from .utils import convert_data_seqs_to_embeddings

from .aprkh_utils.seqio import fetch_seqs

import argparse
import sys 
import h5py

from functools import reduce
from collections import defaultdict


def main():
    args = parse_args()

    if args['batch_size']:
        print(f"setting batch size to {args['batch_size']}...", end=' ')
        set_batch_size(args['batch_size'])
        print("done!")

    print("parsing data")
    sequences = {}
    embeddings = {}
    pairs = []
    # parse kinases
    if args['fkins'].endswith('.h5'):
        embeddings.update(read_embeddings(args['fkins'], args['device']))
    else:
        sequences.update(read_sequences(args['fkins'], to_dict=True))

    # parse substrates
    kinase_keys = set(embeddings.keys()).union(sequences.keys())
    subs_to_site = defaultdict(list)
    with open(args['fsubs'], 'rt') as f:
        for line in f:
            if line.startswith('#'):
                continue 

            #substrate_id, psites = line.split()
            split = line.split()
            substrate_id = split[0]
            # handle either comma or space-separated psites
            psites = []
            for psite in split[1:]:
                psites.extend([int(s.strip()) for s in psite.split(',')])

            subs_to_site[substrate_id].extend(psites)

    # now fetch sequences 
    if args['db']:
        seqs = read_sequences(args['db'], to_dict=True)
    else:
        seqs = fetch_seqs(db='protein', id=[substrate_id])

    # make pairs 
    for sid in subs_to_site.keys():
        if sid not in seqs.keys(): 
            print("WARNING: invalid sequence id:", substrate_id, file=sys.stderr)
            continue

        # add pairs 
        for key in kinase_keys:
            for psite in subs_to_site[sid]:
                pairs.append((key, sid, psite))

        # add sequence 
        sequences[sid] = seqs[sid]

    pairs, embeds2 = convert_data_seqs_to_embeddings(sequences, pairs, args['device'], testing=False)
    embeddings.update(embeds2)


    print("running model...", end=' ')
    results = run_kolossus(pairs, embeds=embeddings, **args['kolossus'])
    print("done!")

    output_results(results, **args['output'])


def output_results(results, dists_fname_out, projections_fname_out):
    if projections_fname_out:
        results, projections = results
        print("writing embeddings to:", projections_fname_out)
        with h5py.File(projections_fname_out, 'w') as fout:
            for _id, e in projections.items():
                fout[_id] = e

    print("writing predictions to", dists_fname_out)
    with open(dists_fname_out, 'wt') as fout:
        print('#kinase\tsubstrate\tpredicted_prob', file=fout)
        for pair, prob in results.items():
            pair = tuple(map(str, pair))
            prob = str(prob)
            print('\t'.join((*pair, prob)), file=fout)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('-s', '--substrates', type=str, required=True, 
                        help="format: \t<substrate_id>\t<psite1>,<psite2>,...")
    parser.add_argument('-k', '--kinases', type=str, required=True, 
                        help='kinase sequences either seqs or embeddings (.h5) must be provided')
    parser.add_argument('--db', type=str, default=None, 
                        help='database of substrate sequences, in fasta format')
    parser.add_argument('--dtype', type=str, default='', 
                        help='data type of sequence embeddings (usually float32)')
    parser.add_argument('-r', '--projections', type=str, default='', 
                        help='name of .h5 files for kolossus projections')
    parser.add_argument('-d', '--device', type=str, default='cpu', 
                        help='default device on which to run model')
    parser.add_argument('--batch_size', type=int, default=None, 
                        help='Number of pairs at a time on which to run model')
    parser.add_argument('-o', '--output', type=str, required=True, 
                        help='desired file path for output')

    args = parser.parse_args()

    real_args = {'kolossus': {'dtype': args.dtype,
                              'device': int(args.device) if args.device.isdigit() else args.device,
                              'return_projections': len(args.projections) > 0},
                 'fsubs': args.substrates,
                 'fkins': args.kinases,
                 'db': args.db,
                 'device': int(args.device) if args.device.isdigit() else args.device,
                 'dtype': args.dtype,
                 'batch_size': args.batch_size,
                 'output': {'dists_fname_out': args.output,
                            'projections_fname_out': args.projections if len(args.projections) > 0 else None}}
    
    return real_args    


if __name__ == '__main__':
    main()




# s is string, i is index for substitution
# w is width (assume odd window), pad is padding character 
# assume odd w 
def extract_window(s, i, w, pad):
    nchars = w // 2
    prefix = [None for _ in range(nchars)]
    suffix = [None for _ in range(nchars)]
    for j in range(nchars):
        # handle prefix 
        if i - j - 1 < 0:
            prefix[-j - 1] = pad
        else:
            prefix[-j - 1] = s[i - j - 1]

        # handle suffix 
        if i + j + 1 >= len(s):
            suffix[j] = pad
        else:
            suffix[j] = s[i + j + 1]

    return ''.join(prefix + [s[i]] + suffix)
            

def chop(s, n):
    i = 0
    out = []
    while i < len(s):
        out.append(s[i:i+n])
        i += n
    return out




import torch
from torch.utils.data import DataLoader
from torch import nn


class Cosine(nn.Module):
    def forward(self, kinases, site):
        return nn.CosineSimilarity()(kinases, site)

# Create Model
class Network(nn.Module):
    def __init__(self):
        super(Network,self).__init__()
        self.kinases_projector1 = nn.Sequential(nn.Linear(5120, 2048), nn.ReLU())
        nn.init.xavier_normal_(self.kinases_projector1[0].weight)
        
        self.kinases_projector2 = nn.Sequential(nn.Linear(2048, 1024), nn.ReLU())
        nn.init.xavier_normal_(self.kinases_projector2[0].weight)

        self.site_projector1 = nn.Sequential(nn.Linear(5120, 2048), nn.ReLU())
        nn.init.xavier_normal_(self.site_projector1[0].weight)
        
        self.site_projector2 = nn.Sequential(nn.Linear(2048, 1024), nn.ReLU())
        nn.init.xavier_normal_(self.site_projector2[0].weight)
        
        self.activator = Cosine()
        
    def forward(self, embedd):
          kinases = embedd[:, :5120]
          site = embedd[:, 5120:]  
          return self.classify(kinases, site)

    def classify(self, kinases, site):
        # get kinases and site projection
        site_projection, kinases_projection = self.project(kinases, site)

        # get cosine similarity (order doesnt' matter since cosine 
        # similarity is symmetric)
        distance = self.activator(kinases_projection, site_projection)

        return distance.squeeze(), site_projection, kinases_projection

    def project(self, kinases, site):
        # this is actually the site projection
        kinases_projection = self.kinases_projector1(kinases)
        kinases_projection = self.kinases_projector2(kinases_projection)

        # this is actually the kinase projection
        site_projection = self.site_projector1(site)
        site_projection = self.site_projector2(site_projection)

        # will actually return (projection of kinase, projection of site)
        return site_projection, kinases_projection

